# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# #**** END GPL LICENSE BLOCK #****

# <pep8 compliant>

import bpy
import bmesh
import subprocess
import os
import sys
import time
from math import atan, pi, degrees, sqrt
import re
import random


def Transform(pat,tabWrite):
    if pat.tex_coords != "UV":
        orient = pat.warp_orientation
        exp = pat.warp_dist_exp
        if pat.warp_types == "REPEAT":
            if pat.warp_repeat_x:
                offset=""
                if pat.rx_offset_value != 0:
                    offset=" offset %s*%.4g"%(pat.rx_offset,pat.rx_offset_value)
                flip=""
                x,y,z=0,0,0
                if pat.rx_flip_x or pat.rx_flip_y or pat.rx_flip_z:
                    if pat.rx_flip_x:
                        x=1
                    if pat.rx_flip_y:
                        y=1
                    if pat.rx_flip_z:
                        z=1
                    flip=" flip <%s,%s,%s>"%(x,y,z)
                tabWrite("warp {repeat x*%.4g%s%s}\n"%(pat.rx_value,offset,flip))
            if pat.warp_repeat_y:
                offset=""
                if pat.ry_offset_value != 0:
                    offset=" offset %s*%.4g"%(pat.ry_offset,pat.ry_offset_value)
                flip=""
                x,y,z=0,0,0
                if pat.ry_flip_x or pat.ry_flip_y or pat.ry_flip_z:
                    if pat.ry_flip_x:
                        x=1
                    if pat.ry_flip_y:
                        y=1
                    if pat.ry_flip_z:
                        z=1
                    flip=" flip <%s,%s,%s>"%(x,y,z)
                tabWrite("warp {repeat x*%.4g%s%s}\n"%(pat.ry_value,offset,flip))
            if pat.warp_repeat_z:
                offset=""
                if pat.rz_offset_value != 0:
                    offset=" offset %s*%.4g"%(pat.rz_offset,pat.rz_offset_value)
                flip=""
                x,y,z=0,0,0
                if pat.rz_flip_x or pat.rz_flip_y or pat.rz_flip_z:
                    if pat.rz_flip_x:
                        x=1
                    if pat.rz_flip_y:
                        y=1
                    if pat.rz_flip_z:
                        z=1
                    flip=" flip <%s,%s,%s>"%(x,y,z)
                tabWrite("warp {repeat x*%.4g%s%s}\n"%(pat.rz_value,offset,flip))
        if pat.warp_types == "CUBIC":
            tabWrite("warp {cubic}\n")
        if pat.warp_types == "PLANAR":
            tabWrite("warp {planar %s %.4g}\n"%(orient,exp))
        if pat.warp_types == "SPHERICAL":
            tabWrite("warp {spherical orientation %s dist_exp %.4g}\n"%(orient,exp)) 
        if pat.warp_types == "CYLINDRICAL":
            tabWrite("warp {cylindrical orientation %s dist_exp %.4g}\n"%(orient,exp)) 
        if pat.warp_types == "TOROIDAL":
            radius = pat.warp_tor_major_radius
            tabWrite("warp {toroidal orientation %s dist_exp %.4g major_radius %.4g}\n"%(orient,exp,radius)) 
    x=pat.warp_turbulence_x
    y=pat.warp_turbulence_y
    z=pat.warp_turbulence_z
    if x != 0 or y != 0 or z != 0:
        tabWrite("turbulence <%.4g,%.4g,%.4g> \n"%(x,y,z))
        tabWrite("octaves %s \n"%pat.modifier_octaves)
        tabWrite("lambda %.4g \n"%pat.modifier_lambda)
        tabWrite("omega %.4g \n"%pat.modifier_omega)
    if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'} :
        tabWrite("frequency %.4g \n"%pat.modifier_frequency)
    if pat.modifier_phase > 0: 
        tabWrite("phase %.4g \n"%pat.modifier_phase) 
    rx,ry,rz = pat.tex_rot_x,pat.tex_rot_y,pat.tex_rot_z
    if rx != 0 or ry != 0 or rz != 0:
        tabWrite("rotate <%.4g,%.4g,%.4g> \n"%(rx, ry, rz))
    mx,my,mz = pat.tex_mov_x,pat.tex_mov_y,pat.tex_mov_z
    if mx != 0 or my != 0 or mz != 0:
        tabWrite("translate <%.4g,%.4g,%.4g> \n"%(mx, my, mz))
    sx,sy,sz = pat.tex_scale_x,pat.tex_scale_y,pat.tex_scale_z
    if sx != 1 or sy != 1 or sz != 1:
        tabWrite("scale <%.4g,%.4g,%.4g> \n"%(1.0/sx, 1.0/sy, 1.0/sz))

def PovFunctions(pat,tabWrite,file):
    if pat.colored_patterns == 'function' and pat.func_list != "NONE":
        tabWrite("{ %s"%pat.func_list)
        file.write("(x")
        if pat.func_plus_x != "NONE":
            if pat.func_plus_x =='increase':
                file.write("*")
            if pat.func_plus_x =='plus':
                file.write("+")
            file.write("%.4g"%pat.func_x)
        file.write(",y")
        if pat.func_plus_y != "NONE":
            if pat.func_plus_y =='increase':
                file.write("*")
            if pat.func_plus_y =='plus':
                file.write("+")
            file.write("%.4g"%pat.func_y)
        file.write(",z")
        if pat.func_plus_z != "NONE":
            if pat.func_plus_z =='increase':
                file.write("*")
            if pat.func_plus_z =='plus':
                file.write("+")
            file.write("%.4g"%pat.func_z)
        sort = -1
        if pat.func_list in {"f_comma","f_crossed_trough","f_cubic_saddle","f_cushion","f_devils_curve",
                             "f_enneper","f_glob","f_heart","f_hex_x","f_hex_y","f_hunt_surface",
                             "f_klein_bottle","f_kummer_surface_v1","f_lemniscate_of_gerono","f_mitre",
                             "f_nodal_cubic","f_noise_generator","f_odd","f_paraboloid","f_pillow",
                             "f_piriform","f_quantum","f_quartic_paraboloid","f_quartic_saddle",
                             "f_sphere","f_steiners_roman","f_torus_gumdrop","f_umbrella"}:
            sort = 0
        if pat.func_list in {"f_bicorn","f_bifolia","f_boy_surface","f_superellipsoid","f_torus"}:
            sort = 1
        if pat.func_list in {"f_ellipsoid","f_folium_surface","f_hyperbolic_torus",
                             "f_kampyle_of_eudoxus","f_parabolic_torus","f_quartic_cylinder","f_torus2"}:
            sort = 2
        if pat.func_list in {"f_blob2","f_cross_ellipsoids","f_flange_cover","f_isect_ellipsoids",
                             "f_kummer_surface_v2","f_ovals_of_cassini","f_rounded_box","f_spikes_2d","f_strophoid"}:
            sort = 3
        if pat.func_list in {"f_algbr_cyl1","f_algbr_cyl2","f_algbr_cyl3","f_algbr_cyl4","f_blob","f_mesh1","f_poly4","f_spikes"}:
            sort = 4
        if pat.func_list in {"f_devils_curve_2d","f_dupin_cyclid","f_folium_surface_2d","f_hetero_mf","f_kampyle_of_eudoxus_2d",
                             "f_lemniscate_of_gerono_2d","f_polytubes","f_ridge","f_ridged_mf","f_spiral","f_witch_of_agnesi"}:
            sort = 5
        if pat.func_list in {"f_helix1","f_helix2","f_piriform_2d","f_strophoid_2d"}:
            sort = 6
        if pat.func_list == "f_helical_torus":
            sort = 7
        if sort > -1:
            file.write(",%.4g"%pat.func_P0)
        if sort > 0:
            file.write(",%.4g"%pat.func_P1)
        if sort > 1:
            file.write(",%.4g"%pat.func_P2)
        if sort > 2:
            file.write(",%.4g"%pat.func_P3)
        if sort > 3:
            file.write(",%.4g"%pat.func_P4)
        if sort > 4:
            file.write(",%.4g"%pat.func_P5)
        if sort > 5:
            file.write(",%.4g"%pat.func_P6)
        if sort > 6:
            file.write(",%.4g"%pat.func_P7)
            file.write(",%.4g"%pat.func_P8)
            file.write(",%.4g"%pat.func_P9)
        file.write(")}\n")

def PovPatterns(pat,tabWrite):
    tabWrite("%s\n"%pat.colored_patterns)
    if pat.tex_as_normal and pat.tex_use_slope == False and pat.tex_nor_color_map == False:
        tabWrite("%.4g\n"%pat.tex_nor)
    if pat.colored_patterns == 'agate': 
        tabWrite("agate_turb %.4g\n"%pat.agate_turbulence)                            
    if pat.colored_patterns == 'crackle':
        crack="form <%.4g,%.4g,%.4g>"%(pat.crackle_form_x,pat.crackle_form_y,pat.crackle_form_z)
        crack+=" metric %.4g"%pat.crackle_metric
        if pat.crackle_solid:
            crack+=" solid"
        tabWrite("%s\n"%crack)
    if pat.colored_patterns in {'spiral1', 'spiral2', 'tiling'}: 
        tabWrite("%s\n"%pat.modifier_numbers)
    if pat.colored_patterns == 'quilted': 
        tabWrite("control0 %s control1 %s\n"%(pat.modifier_control0, pat.modifier_control1))                            
    if pat.colored_patterns == 'mandel': 
        tabWrite("%s exponent %s \n"%(pat.f_iter, pat.f_exponent))  
    if pat.colored_patterns == 'julia': 
        tabWrite("<%.4g, %.4g> %s exponent %s \n"%(pat.julia_complex_1, pat.julia_complex_2, pat.f_iter, pat.f_exponent))   
    if pat.colored_patterns == 'magnet' and pat.magnet_style == 'mandel': 
        tabWrite("%s mandel %s \n"%(pat.magnet_type, pat.f_iter))  
    if pat.colored_patterns == 'magnet' and pat.magnet_style == 'julia':  
        tabWrite("%s julia <%.4g, %.4g> %s\n"%(pat.magnet_type, pat.julia_complex_1, pat.julia_complex_2, pat.f_iter)) 
    if pat.colored_patterns in {'mandel', 'julia', 'magnet'}:
        tabWrite("interior %s, %.4g\n"%(pat.f_ior, pat.f_ior_fac)) 
        tabWrite("exterior %s, %.4g\n"%(pat.f_eor, pat.f_eor_fac))
    if pat.colored_patterns == 'gradient': 
        tabWrite("<%s, %s, %s> \n"%(pat.grad_orient_x, pat.grad_orient_y, pat.grad_orient_z))
    if pat.colored_patterns == 'pavement':
        numTiles=pat.pave_tiles
        numPattern=1
        if pat.pave_sides == '4' and pat.pave_tiles == 3: 
             numPattern = pat.pave_pat_2
        if pat.pave_sides == '6' and pat.pave_tiles == 3: 
            numPattern = pat.pave_pat_3
        if pat.pave_sides == '3' and pat.pave_tiles == 4: 
            numPattern = pat.pave_pat_3
        if pat.pave_sides == '3' and pat.pave_tiles == 5: 
            numPattern = pat.pave_pat_4
        if pat.pave_sides == '4' and pat.pave_tiles == 4: 
            numPattern = pat.pave_pat_5
        if pat.pave_sides == '6' and pat.pave_tiles == 4: 
            numPattern = pat.pave_pat_7
        if pat.pave_sides == '4' and pat.pave_tiles == 5: 
            numPattern = pat.pave_pat_12
        if pat.pave_sides == '3' and pat.pave_tiles == 6: 
            numPattern = pat.pave_pat_12
        if pat.pave_sides == '6' and pat.pave_tiles == 5: 
            numPattern = pat.pave_pat_22
        if pat.pave_sides == '4' and pat.pave_tiles == 6: 
            numPattern = pat.pave_pat_35
        if pat.pave_sides == '6' and pat.pave_tiles == 6: 
            numTiles = 5                                
        tabWrite("number_of_sides %s number_of_tiles %s pattern %s form %s \n"%(pat.pave_sides, numTiles, numPattern, pat.pave_form))

def PovMaps(pat,tabWrite,texorpig,pos,posColor,numColor):
    if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'}:
        tabWrite("[%.4g %s]\n"%(pos,posColor))
    if (pat.tex_use_pmap or pat.tex_as_material) and pat.colored_patterns in {'checker', 'hexagon', 'square', 'triangular', 'brick'}:
        posColor="%s{%s}"%(texorpig,posColor)
    if pat.colored_patterns in {'brick','checker'} and numColor < 3:
        tabWrite("%s\n"%posColor)
    if pat.colored_patterns == 'hexagon' and numColor < 4 :
        tabWrite("%s\n"%posColor)
    if pat.colored_patterns == 'square' and numColor < 5 :
        tabWrite("%s\n"%posColor)
    if pat.colored_patterns == 'triangular' and numColor < 7 :
        tabWrite("%s\n"%posColor)

def imgMapBG(wts):
    image_mapBG = ""
    if wts.texture_coords == 'VIEW':
        image_mapBG = " map_type 0 "
    elif wts.texture_coords == 'ANGMAP':
        image_mapBG = " map_type 1 "
    elif wts.texture_coords == 'TUBE':
        image_mapBG = " map_type 2 "

    if wts.texture.use_interpolation:
        image_mapBG += " interpolate 2 "
    if wts.texture.extension == 'CLIP':
        image_mapBG += " once "

    return image_mapBG


def path_image(image):
    return bpy.path.abspath(image.filepath, library=image.library)

def string_strip_hyphen(name):
    return name.replace("-", "")


def safety(name, Level):
    try:
        if int(name) > 0:
            prefix = "shader"
    except:
        prefix = ""
    prefix = "shader_"
    name = string_strip_hyphen(name)
    if Level == 2:
        return prefix + name
    elif Level == 1:
        return prefix + name + "0"
    elif Level == 3:
        return prefix + name + "1" 

def is_renderable(scene, ob):
    return (ob.is_visible(scene) and not ob.hide_render)


def renderable_objects(scene):
    return [ob for ob in scene.objects if is_renderable(scene, ob)]


tabLevel = 0


def write_pov(filename, scene, info_callback=None):
    import mathutils
    file = open(filename, "w")

    scene = bpy.context.scene

    render = scene.render
    world = scene.world
    global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')

    def setTab(tabtype, spaces):
        TabStr = ""
        if tabtype == 'NONE':
            TabStr = ""
        elif tabtype == 'TAB':
            TabStr = "\t"
        elif tabtype == 'SPACE':
            TabStr = spaces * " "
        return TabStr

    tab = setTab(scene.pov.indentation_character, scene.pov.indentation_spaces)

    def tabWrite(str_o):
        if not scene.pov.tempfiles_enable:
            global tabLevel
            brackets = str_o.count("{") - str_o.count("}") + str_o.count("[") - str_o.count("]")
            if brackets < 0:
                tabLevel = tabLevel + brackets
            if tabLevel < 0:
                print("Indentation Warning: tabLevel = %s" % tabLevel)
                tabLevel = 0
            if tabLevel >= 1:
                file.write("%s" % tab * tabLevel)
            file.write(str_o)
            if brackets > 0:
                tabLevel = tabLevel + brackets
        else:
            file.write(str_o)

    def uniqueName(name, nameSeq):

        if name not in nameSeq:
            name = string_strip_hyphen(name)
            return name

        name_orig = name
        i = 1
        while name in nameSeq:
            name = "%s_%.3d" % (name_orig, i)
            i += 1
        name = string_strip_hyphen(name)
        return name

    def writeMatrix(matrix):
        tabWrite("matrix <%.6f, %.6f, %.6f,  %.6f, %.6f, %.6f,  %.6f, %.6f, %.6f,  %.6f, %.6f, %.6f>\n" %
                 (matrix[0][0], matrix[1][0], matrix[2][0],
                  matrix[0][1], matrix[1][1], matrix[2][1],
                  matrix[0][2], matrix[1][2], matrix[2][2],
                  matrix[0][3], matrix[1][3], matrix[2][3]))

    def MatrixAsPovString(matrix):
        sMatrix = ("matrix <%.6f, %.6f, %.6f,  %.6f, %.6f, %.6f,  %.6f, %.6f, %.6f,  %.6f, %.6f, %.6f>\n" %
                   (matrix[0][0], matrix[1][0], matrix[2][0],
                    matrix[0][1], matrix[1][1], matrix[2][1],
                    matrix[0][2], matrix[1][2], matrix[2][2],
                    matrix[0][3], matrix[1][3], matrix[2][3]))
        return sMatrix

    def mesh_triangulate(me):
        import bmesh
        bm = bmesh.new()
        bm.from_mesh(me)
        bmesh.ops.triangulate(bm, faces=bm.faces)
        bm.to_mesh(me)
        bm.free()

    def declareFinPigNor(material):
        povName = string_strip_hyphen(bpy.path.clean_name(ob.name))
        povMatName = string_strip_hyphen(bpy.path.clean_name(material.name))
        if material and material.type == "SURFACE" and material.pov.povray_finish:
            mat = material.pov
            tabWrite("#declare %s_finish = finish {\n" %povMatName)
            if mat.m_use_brilliance :
                tabWrite("brilliance %.3g\n" % (mat.m_brilliance))
            if mat.m_use_crand and mat.m_crand != 0:
                tabWrite("crand %.3g\n" % (mat.m_crand))
            if mat.m_use_spec:
                if mat.spec_use_phong == False:
                    tabWrite("specular")
                    if mat.spec_use_albedo:
                        file.write(" albedo")
                    file.write(" %.3g\n" % (material.specular_intensity))
                    tabWrite("roughness %.3g\n" % mat.spec_size)
                if mat.spec_use_phong:
                    tabWrite("phong")
                    if material.pov.spec_use_albedo:
                        file.write(" albedo")
                    file.write(" %.3g\n" % (material.specular_intensity))
                    tabWrite("phong_size %.3g\n" % (material.specular_hardness/2))
                if mat.m_use_specmetallic:
                    tabWrite("metallic %.3g\n" %mat.m_specmetallic)
            if mat.m_use_diffuse:
                albedo=""
                if mat.m_diffuse_albedo:
                    albedo="albedo"
                tabWrite("diffuse %s %.3g\n" % (albedo,material.diffuse_intensity))
            if mat.m_use_ambient:
                tabWrite("ambient <%.3g, %.3g, %.3g>\n" % material.pov.m_ambient_color[:])
            if mat.m_use_emission:
                tabWrite("emission %.3g\n" % material.emit)  # New in POV-Ray 3.7
            if material.pov.conserve_energy:
                tabWrite("conserve_energy\n")
            if material.raytrace_mirror.use:
                raytrace_mirror = material.raytrace_mirror
                if raytrace_mirror.reflect_factor:
                    tabWrite("reflection {\n")
                    tabWrite("<%.3g,%.3g,%.3g>" % material.mirror_color[:])
                    file.write(" * %.3g\n"%(material.raytrace_mirror.reflect_factor))
                    if material.pov.m_metallic > 0:
                        tabWrite("metallic %.3g\n" % (material.pov.m_metallic))
                    if material.pov.m_reflect_fresnel:
                        tabWrite("fresnel on\n")
                        tabWrite("falloff %.3g exponent %.3g " % (raytrace_mirror.fresnel, raytrace_mirror.fresnel_factor))
                    tabWrite("}\n")
            if scene.pov.sss_enable and mat.sss_mat_enable:
                    subsurface_scattering = material.subsurface_scattering
                    tabWrite("subsurface { translucency <%.3g, %.3g, %.3g> }\n" %material.pov.sss_translucency_color[:])
            if mat.irid_enable:
                tabWrite("irid { %.4g thickness %.4g turbulence %.4g }" % \
                    (material.pov.irid_amount, material.pov.irid_thickness,
                    material.pov.irid_turbulence))
            tabWrite("}\n\n")
        if material.type=='VOLUME':
            povTexName = string_strip_hyphen(bpy.path.clean_name(material.name))+"_tex"
            tabWrite("#declare %s = texture {pigment {color rgbt 1}}\n"%povTexName)
        if material.type=='SURFACE':
            povTexName = string_strip_hyphen(bpy.path.clean_name(material.name))+"_tex"
            diffuse_color = material.diffuse_color[:]
            r,g,b = diffuse_color[0], diffuse_color[1], diffuse_color[2]
            povTransmit=material.pov.m_transmit
            povFilter = material.pov.m_filter
            imageDif = 0            
            for t in material.texture_slots:
                image_filename = None
                image_blend = False
    
                if t and t.texture.pov.tex_type == 'IMAGE' and t.use and t.texture.pov.tex_image_path:
                    image_filename = t.texture.pov.tex_image_path
                if t and t.texture.pov.tex_type == 'PATTERN' and t.use:
                    image_blend = True
                if image_filename or image_blend:
    
                    if t.texture.pov.tex_as_pigment and (image_filename or image_blend):
                        pat = t.texture.pov
                        imageDif += 1

    
            for t in material.texture_slots:
                image_filename = None
                image_blend = False
    
                if t and t.texture.pov.tex_type == 'IMAGE' and t.use and t.texture.pov.tex_image_path:
                    image_filename = t.texture.pov.tex_image_path
                if t and t.texture.pov.tex_type == 'PATTERN':
                    image_blend = True
                if image_filename or image_blend:
                    pat = t.texture.pov
                    if pat.tex_as_pigment or pat.tex_as_ppattern:
                        povTexName = string_strip_hyphen(bpy.path.clean_name(t.texture.name))
                        entry = "#declare %s = pigment{"%povTexName
                        if pat.tex_coords == "UV":
                            entry+=" uv_mapping"
                        tabWrite("%s\n"%entry)
                        if image_filename:
                            tabWrite("image_map {\n")
                            tabWrite('"%s" \n'%image_filename)
                            tabWrite("map_type 0 interpolate 2 filter all %.4g transmit all %.4g }\n"%(pat.tex_pigment_filter,pat.tex_pigment_transmit))
                            Transform(pat,tabWrite)
                            tabWrite("}\n")
                        if image_blend:
                            PovPatterns(pat,tabWrite)
                            PovFunctions(pat,tabWrite,file)
                            if pat.tex_as_pigment or pat.tex_as_ppattern:
                                texorpig = "pigment"
                                if pat.tex_use_pmap:
                                    if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'} :
                                        tabWrite("pigment_map {\n")
                                else:
                                    if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'} :
                                        tabWrite("color_map {\n")
                                numColor=0
                                numEl=-1
                                for el in t.texture.color_ramp.elements:
                                    numColor+=1
                                    numEl+=1
                                    pos = el.position
                                    col=el.color
                                    colR,colG,colB,colA = col[0],col[1],col[2],1-col[3]
                                    posColor="color rgbf<%.4g,%.4g,%.4g,%.4g>"%(colR,colG,colB,colA)
                                    if pat.tex_use_pmap:
                                        for entry in t.texture.povray_pigment:
                                            if entry.pigment_number == numEl:
                                                if entry.pigment_name != "":
                                                    povPigName=string_strip_hyphen(bpy.path.clean_name(entry.pigment_name))
                                                    posColor="%s"%povPigName
                                                else:
                                                    posColor="color rgbft<%.4g,%.4g,%.4g,%.4g,%.4g>"%(colR,colG,colB,colA,entry.color_transmit)
                                    PovMaps(pat,tabWrite,texorpig,pos,posColor,numColor)
                                if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'} :
                                    tabWrite("} \n")
                            if pat.colored_patterns == 'brick':
                                tabWrite("brick_size <%.4g, %.4g, %.4g> mortar %.4g \n"%(pat.brick_size_x, pat.brick_size_y, pat.brick_size_z, pat.brick_mortar))
                            Transform(pat,tabWrite)
                            tabWrite("}\n")
            imageNor = 0
            for t in material.texture_slots:
                image_filename = None
                image_blend = False
    
                if t and t.texture.pov.nor_type == 'IMAGE' and t.use and t.texture.pov.tex_image_path:
                    image_filename = t.texture.pov.tex_image_path
                if t and t.texture.pov.nor_type == 'PATTERN' and t.use:
                    image_blend = True
                if image_filename or image_blend:
                    if t.texture.pov.tex_as_normal:
                        povTexName = string_strip_hyphen(bpy.path.clean_name(t.texture.name))
    
                    if t.texture.pov.tex_as_normal and image_filename:
                        pat = t.texture.pov
                        imageNor+=1
                        tabWrite("#declare %s = normal {\n"%povTexName)
                        if t.texture.pov.tex_coords == "UV":
                            tabWrite("uv_mapping bump_map {\n")
                        if t.texture.pov.tex_coords != "UV":
                            tabWrite("bump_map {\n")
                        tabWrite('"%s" \n'%image_filename)
                        if t.texture.pov.tex_coords == "UV":
                            tabWrite("map_type 0 interpolate 2 bump_size %.4g }\n"%pat.tex_nor)
                        if t.texture.pov.tex_coords != "UV":
                            tabWrite("bump_size %.4g }\n"%pat.tex_nor)
                        Transform(pat,tabWrite)
                        tabWrite("}\n")
    
                    if t.texture.pov.tex_as_normal and image_blend:
                        pat = t.texture.pov
                        imageNor+=1
                        if t.texture.pov.tex_coords == "UV":
                            tabWrite("#declare %s = normal { uv_mapping\n"%povTexName)
                        if t.texture.pov.tex_coords != "UV":
                            tabWrite("#declare %s = normal {\n"%povTexName)
                        if pat.tex_nor_color_map:
                            tabWrite("pigment_pattern {\n")
                        PovPatterns(pat,tabWrite)
                        PovFunctions(pat,tabWrite,file)
                        if pat.tex_nor_color_map:
                            if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'}:                        
                                tabWrite("color_map {\n")
                            numColor=0
                            for el in t.texture.color_ramp.elements:
                                numColor+=1
                                pos = el.position
                                col=el.color
                                colR,colG,colB = col[0],col[1],col[2]
                                if t.texture.pov.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'} :
                                    tabWrite("[%.4g color rgb<%.4g,%.4g,%.4g>] \n"%(pos,colR,colG,colB))
                                if t.texture.pov.colored_patterns in {'brick','checker'} and numColor < 3:
                                    tabWrite("color rgb<%.4g,%.4g,%.4g> \n"%(colR,colG,colB))
                                if t.texture.pov.colored_patterns == 'hexagon' and numColor < 4 :
                                    tabWrite("color rgb<%.4g,%.4g,%.4g> \n"%(colR,colG,colB))
                                if t.texture.pov.colored_patterns == 'square' and numColor < 5 :
                                    tabWrite("color rgb<%.4g,%.4g,%.4g> \n"%(colR,colG,colB))
                                if t.texture.pov.colored_patterns == 'triangular' and numColor < 7 :
                                    tabWrite("color rgb<%.4g,%.4g,%.4g> \n"%(colR,colG,colB))
    
                            if t.texture.pov.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'} :                        
                                tabWrite("} \n")
                            if t.texture.pov.colored_patterns == 'brick':                        
                                tabWrite("brick_size <%.4g, %.4g, %.4g > mortar %s \n"%(t.texture.pov.brick_size_x,t.texture.pov.brick_size_y,t.texture.pov.brick_size_z,t.texture.pov.brick_mortar))
                        if pat.tex_use_slope:
                            tabWrite("slope_map {\n")
                            for entry in t.texture.povray_slope:
                                tabWrite("[%s <%s,%s>]\n"%(entry.slope_value,entry.slope_height,entry.slope_slope))
                            tabWrite("}\n")
                            tabWrite("accuracy %s\n"%pat.tex_slope_accuracy)
                        Transform(pat,tabWrite)
                        tabWrite("}\n")
                        if pat.tex_nor_color_map:
                            tabWrite(" %.4g }\n"%pat.tex_nor)
    
            if material and material.type == "VOLUME":
                tabWrite("#declare Texture_1_%s = texture {pigment {color rgbt 1}}\n"%povName)

    def exportCamera():
        camera = scene.camera
        matrix = global_matrix * camera.matrix_world
        if camera.data.dof_object:
            focus_matrix = global_matrix * camera.data.dof_object.matrix_world
        Qsize = (float(render.resolution_x) / float(render.resolution_y)) - camera.data.pov.correction_aspect_ratio



        if camera.data.pov.cam_type == 'perspective' and camera.data.pov.cam_architectural:
            tabWrite("#declare Camera_Location = <%.6f, %.6f, %.6f>;\n" % matrix.translation[:])
            if camera.constraints and camera.constraints["Track To"]:
                matrixConstrains = global_matrix * camera.constraints["Track To"].target.matrix_world
                tabWrite("#declare Camera_Look_At = <%.6f, %.6f, %.6f>;\n" % matrixConstrains.translation[:])
            else:
                tabWrite("#declare Camera_Look_At = <0, 0, -1>;\n")
            tabWrite("#declare Camera_Angle = %f;\n" % (360.0 * atan(16.0 / camera.data.lens) / pi))
            tabWrite('#include "transforms.inc"\n')
            tabWrite("#declare Cam_V = Camera_Look_At - Camera_Location;\n")
            tabWrite("#declare Cam_Ho = sqrt(pow(Cam_V.x,2)+pow(Cam_V.z ,2));\n")
            tabWrite("#declare Cam_Y = Camera_Look_At.y - Camera_Location.y;\n")
            tabWrite("camera {\n")
            tabWrite("angle  Camera_Angle\n")
            tabWrite("right <%s, 0, 0>\n" % -Qsize)
            tabWrite("location  <0,Camera_Look_At.y,-Cam_Ho>\n")
            tabWrite("matrix<1,0,0, 0,1,0, 0,Cam_Y/Cam_Ho,1, 0,0,0>\n")
            tabWrite("Reorient_Trans(z,<Cam_V.x,0,Cam_V.z>)\n")
            tabWrite("translate<Camera_Look_At.x+2,1.25,Camera_Look_At.z+2>\n")
        else:
            tabWrite("camera {\n")
            if camera.data.pov.cam_type not in {'cylindrical', 'mesh_camera','perspective'}:
                tabWrite("%s\n"%camera.data.pov.cam_type)
            if camera.data.pov.cam_type == 'cylindrical': 
                tabWrite("cylinder %s\n"%camera.data.pov.cylinder_type)
            if camera.data.pov.cam_type == 'mesh_camera':
                tabWrite("mesh_camera{ %.6f %s\n"%(camera.data.pov.rays_per_pixel, camera.data.pov.mesh_cam_distr))
                meshName = string_strip_hyphen(bpy.path.clean_name(camera.data.pov.mesh_cam_obj))
                tabWrite("mesh{%s}\n" %(meshName))
                tabWrite("}\n")
                tabWrite("location <0,0,.01>")
                tabWrite("direction <0,0,-1>")
            else:
                tabWrite("location  <0, 0, 0>\n")
                tabWrite("look_at  <0, 0, -1>\n")
                tabWrite("right <%s, 0, 0>\n" % - Qsize)
                tabWrite("up <0, 1, 0>\n")
                tabWrite("angle  %f\n" % (360.0 * atan(16.0 / camera.data.lens) / pi))
    
                tabWrite("rotate  <%.6f, %.6f, %.6f>\n" % \
                         tuple([degrees(e) for e in matrix.to_3x3().to_euler()]))
                tabWrite("translate <%.6f, %.6f, %.6f>\n" % matrix.translation[:])
        if camera.data.pov.dof_enable :
            tabWrite("aperture %.3g\n" % camera.data.pov.dof_aperture)
            tabWrite("blur_samples %d %d\n" % \
                     (camera.data.pov.dof_samples_min, camera.data.pov.dof_samples_max))
            tabWrite("variance 1/%d\n" % camera.data.pov.dof_variance)
            tabWrite("confidence %.3g\n" % camera.data.pov.dof_confidence)
            if camera.data.dof_object:
                tabWrite("focal_point <%.6f, %.6f, %.6f>\n" % focus_matrix.translation[:])
        tabWrite("}\n")

    def exportLamps(lamps):
        global lampCount
        lampCount = 0
        for ob in lamps:
            lamp = ob.data
            matrix = global_matrix * ob.matrix_world
            color = tuple([c * lamp.energy for c in lamp.color])
            tabWrite("light_source {\n")
            tabWrite("<%.3g,%.3g,%.3g>\n"%matrix.translation[:])
            tabWrite("color rgb<%.3g, %.3g, %.3g>\n" % color)

            if lamp.type == 'POINT':
                pass
            elif lamp.type == 'SPOT':
                tabWrite("spotlight\n")
                tabWrite("falloff %.2f\n" % lamp.pov.falloff_spot)
                tabWrite("radius %.6f\n" % (degrees(lamp.spot_size) / 2.0))
                tabWrite("tightness %.3g\n"%(lamp.spot_blend*100)) 
                trackTo = ob.constraints["Track To"].target
                trackToMatrix = global_matrix * trackTo.matrix_world
                tabWrite("point_at  <%.3g, %.3g, %.3g>\n"%(trackToMatrix[0][3], trackToMatrix[1][3], trackToMatrix[2][3]))
            elif lamp.type == 'SUN':
                tabWrite("parallel\n")
                tabWrite("point_at  <0, 0, -1>\n")

            elif lamp.type == 'AREA':
                tabWrite("area_illumination\n")
                tabWrite("fade_distance %.6f\n" % (lamp.distance / 2.0))
                tabWrite("fade_power %d\n" % 2)
                size_x = lamp.size
                samples_x = lamp.shadow_ray_samples_x
                if lamp.shape == 'SQUARE':
                    size_y = size_x
                    samples_y = samples_x
                else:
                    size_y = lamp.size_y
                    samples_y = lamp.shadow_ray_samples_y

                tabWrite("area_light <%.6f,0,0>,<0,%.6f,0> %d, %d\n" % \
                         (size_x, size_y, samples_x, samples_y))
                if lamp.shadow_ray_sample_method == 'CONSTANT_JITTERED':
                    if lamp.use_jitter:
                        tabWrite("jitter\n")
                else:
                    tabWrite("adaptive 1\n")
                    tabWrite("jitter\n")

            if lamp.type == 'HEMI':
                tabWrite("shadowless\n")
            if lamp.type not in {'SUN', 'AREA', 'HEMI','POINT'}:
                tabWrite("fade_distance %.6f\n" % (lamp.distance / 2.0))
                if lamp.falloff_type == 'INVERSE_SQUARE':
                    tabWrite("fade_power %d\n" % 2)
                elif lamp.falloff_type == 'INVERSE_LINEAR':
                    tabWrite("fade_power %d\n" % 1)
                elif lamp.falloff_type == 'CONSTANT':
                    pass
                elif lamp.falloff_type == 'CUSTOM_CURVE':
                    tabWrite("fade_power %d\n" % 4)
            if lamp.pov.media_interaction == False:
                tabWrite("media_interaction off\n")
            tabWrite("}\n")
            lampCount += 1



####################################################################################################

    def exportMeta(metas):
        global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
        if not scene.pov.tempfiles_enable and scene.pov.comments_enable and len(metas) >= 1:
            file.write("//--Blob objects--\n\n")
        childs=[]
        for ob in metas:
            if ob.children:
                for child in ob.children:
                    childs.append(child.name)
        for ob in metas:
            if ob.name not in childs:
                matrix=global_matrix*ob.matrix_world
                dS=ob.scale
                scale="scale <%.4g,%.4g,%.4g>"%(dS[0],dS[2],dS[1])
                loc = matrix.translation[:]
                trans="translate <%.6f, %.6f, %.6f>"%loc
                meta = ob.data
                elements = [elem for elem in meta.elements if elem.type in {'BALL', 'ELLIPSOID'}]
                if elements:
                    tabWrite("blob {\n")
                    tabWrite("threshold %.4g\n" % meta.threshold)
                    for elem in elements:
                        loc = global_matrix*ob.matrix_world*elem.co
                        stiffness = elem.stiffness
                        if elem.use_negative:
                            stiffness = - stiffness
                        if elem.type == 'BALL':
                            tabWrite("sphere { <0,0,0>, %.4g, %.4g %s %s}\n" % \
                                     (elem.radius, stiffness, scale, trans))
                if ob.children:
                    for child in ob.children:
                        for obC in metas:
                            if child.name==obC.name:
                                tex=""
                                if len(obC.material_slots)>0:
                                    povCName = string_strip_hyphen(bpy.path.clean_name(obC.name))
                                    tex="texture {%s_tex}"%povCName
                                matrixC=global_matrix*obC.matrix_world
                                dSC=obC.scale
                                scaleC="scale <%.4g,%.4g,%.4g>"%(dSC[0],dSC[2],dSC[1])
                                rotC=tuple([degrees(e) for e in matrixC.to_3x3().to_euler()])
                                rotXC=rotC[0]+90
                                rotateC="rotate <%.4g,%.4g,%.4g>"%(rotXC,rotC[1],rotC[2])
                                meta = obC.data
                                elements = [elem for elem in meta.elements if elem.type in {'BALL', 'ELLIPSOID'}]
                                if elements:
                                    for elem in elements:
                                        locC = matrixC.translation[:]
                                        transC="translate <%.6f, %.6f, %.6f>"%locC
                                        stiffness = elem.stiffness
                                        if elem.use_negative:
                                            stiffness = - stiffness
                                        if elem.type == 'BALL':
                                            tabWrite("sphere { <0,0,0>, %.4g, %.4g %s %s %s %s}\n" % \
                                                     (elem.radius, stiffness, tex, scaleC, rotateC, transC))


                if len(ob.material_slots)>0:
                    povName = string_strip_hyphen(bpy.path.clean_name(ob.name))
                    tabWrite("texture {%s_tex}}\n"%povName)
                else:
                    tabWrite("texture {pigment{rgb 0.8}}}\n")

    def exportObject(ob, povName):
        name=ob.name
        if ob.pov.object_write_type=='BOOLEAN' and ob.pov.boolean_parent==False:
            name=ob.pov.boolean_data
        povName = string_strip_hyphen(bpy.path.clean_name(name))
        obPov="%s_ob"%povName
        if ob.pov.noshadow:
            obPov+=" no_shadow"
        if ob.pov.noreflection:
            obPov+=" no_reflection"
        if ob.pov.noimage:
            obPov+=" no_image"
        if ob.pov.object_write_type=='BOOLEAN':
            global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
            matrix=global_matrix*ob.matrix_world
            dS=ob.scale
            rot=tuple([degrees(e) for e in matrix.to_3x3().to_euler()])
            rotX=rot[0]+90
            dT=matrix.translation[:]
            obPov+=" scale <%.4g,%.4g,%.4g>"%(dS[0],dS[2],dS[1])
            obPov+=" rotate <%.4g,%.4g,%.4g>"%(rotX,rot[1],rot[2])
            obPov+=" translate <%.6f, %.6f, %.6f>"%dT
        tabWrite("object {%s \n"%obPov)
        mats=[]
        for poly in ob.data.polygons:
            if poly.material_index not in mats:
                mats.append(poly.material_index)
        if len(mats) < 2:
            exportObjectTextures(scene, ob)
        exportObjectMaterials(scene, ob, povName)
        tabWrite("}\n")

    def exportObjectTextures(scene, ob):

        if len(ob.material_slots)==0:
            tabWrite("texture {pigment{rgb 0.8}}\n")
        if len(ob.material_slots)>0:
            if ob.active_material:
                am = ob.active_material
                if am.pov.povray_finish == False or (am.pov.povray_finish and am.pov.finish_as_mat):
                    tabWrite("texture{%s_tex}\n"%string_strip_hyphen(bpy.path.clean_name(ob.active_material.name)))

    def exportObjectMaterials(scene, ob, povName):
        if len(ob.material_slots)>0:
            material=ob.material_slots[0].material
            if material and material.type=='VOLUME':
                exportMediaMaterials(material)
            if ob.pov.m_ior > 1:
                tabWrite("interior {\n")
                tabWrite("ior %.6f\n" % ob.pov.m_ior)
                if ob.pov.refraction_caustics:
                    if scene.pov.photon_enable and ob.pov.refraction_photons:
                        tabWrite("dispersion %.6f\n" % ob.pov.photons_dispersion)
                        tabWrite("dispersion_samples %.d\n" % ob.pov.photons_dispersion_samples)
                    else:
                        tabWrite("caustics %.3g\n" % ob.pov.fake_caustics_power)    
                    if ob.pov.fade_distance > 0.000001:
                        tabWrite("fade_distance %.3g\n" % ob.pov.fade_distance)
                        tabWrite("fade_power %.3g\n" % ob.pov.fade_power)
                        tabWrite("fade_color <%.3g, %.3g, %.3g>\n" % ob.pov.fade_color[:])
                tabWrite("}\n")
        if scene.pov.photon_enable:
            reflection = ""
            refraction = ""
            target = ""
            collect = ""
            if ob.pov.reflection_photons:
                reflection = "reflection on"
            if ob.pov.refraction_caustics:
                refraction = "refraction on"
            if ob.pov.target_photons:
                target = "target"
            if ob.pov.collect_photons==False:
                collect = "collect off"
            tabWrite("photons {%s %s %s %s}\n"%(reflection,refraction,target,collect))
        if scene.pov.radio_enable:
            tabWrite("radiosity { importance %.4g }\n"%ob.pov.importance_value)

    def declareTexture(material):
        povName = string_strip_hyphen(bpy.path.clean_name(ob.name))
        povMatName = string_strip_hyphen(bpy.path.clean_name(material.name))
        if material and material.pov.povray_finish and material.pov.finish_as_mat:
            tabWrite("#declare %s_tex =\n" %povMatName)
            diffuse_color = material.diffuse_color[:]
            r,g,b = diffuse_color[0], diffuse_color[1], diffuse_color[2]
            transparent = ""
            transparent_value = ""
            if material.pov.m_transparent:
                transparent="ft"
                povTransmit=material.pov.m_transmit
                povFilter = material.pov.m_filter
                transparent_value=",%.4g,%.4g"%(povFilter,povTransmit)
            povPigName = "color rgb%s <%.4g,%.4g,%.4g%s>"%(transparent,r,g,b,transparent_value)
            for slot in material.texture_slots:
                if slot and slot.texture.name in scene.povray_pigments:
                    povPigName="%s"%string_strip_hyphen(bpy.path.clean_name(slot.texture.name))
            texture="texture {pigment {%s}"%povPigName
            for slot in material.texture_slots:
                if slot and slot.texture.name in scene.povray_normals:
                    texture+=" normal {%s}"%string_strip_hyphen(bpy.path.clean_name(slot.texture.name))
            texture+=" finish{%s_finish}}"%povMatName
            tabWrite("%s\n"%texture)
        if material and material.pov.povray_finish==False:
            if material.pov.povray_mat_type=="DEFAULT":
                tabWrite("#declare %s_tex = texture {pigment{rgb 0.8}}\n" %povMatName)
            if material.pov.povray_mat_type=="LAYERED":
                tabWrite("#declare %s_tex =\n" %povMatName)
                if material.povray_layered_texture:
                    for entry in material.povray_layered_texture:
                        povPigName="rgb 0.8"
                        if entry.layer_pigment != "":
                            povPigName=string_strip_hyphen(bpy.path.clean_name(entry.layer_pigment))
                        texture="texture {pigment{%s}"%povPigName
                        if entry.layer_normal != "":
                            povNorName=string_strip_hyphen(bpy.path.clean_name(entry.layer_normal))
                            texture+=" normal{%s}"%povNorName
                        if entry.layer_finish != "":
                            povFinName=string_strip_hyphen(bpy.path.clean_name(entry.layer_finish))
                            texture+=" finish{%s_finish}"%povFinName
                        texture+="}"
                        tabWrite("%s\n" %texture)
                else:
                    tabWrite("texture {pigment{rgb 0.8}}\n")
            if material.pov.povray_mat_type=="PIGMENTPATTERN":
                if material.pov.m_pp != "":
                    ppattern=string_strip_hyphen(bpy.path.clean_name(material.pov.m_pp))
                else:
                    tabWrite("#declare %s_tmap = pigment {rgb 0}\n" %povMatName)
                    ppattern="%s_tmap"%povMatName
                tabWrite("#declare %s_tex = texture {pigment_pattern {%s}\n" %(povMatName,ppattern))
                tex_1=ppattern
                if material.pov.m_pp_first != "":
                    tex_1=string_strip_hyphen(bpy.path.clean_name(material.pov.m_pp_first))
                tmap="[%.4g %s_tex]"%(material.pov.m_first_value,tex_1)
                if material.pov.m_pp_second != "":
                    tex_2=string_strip_hyphen(bpy.path.clean_name(material.pov.m_pp_second))
                    tmap+="[%.4g %s_tex]" %(material.pov.m_second_value,tex_2)
                tabWrite("texture_map {%s}}\n" %tmap)
            if material.pov.povray_mat_type=="PATTERN":
                if material.texture_slots[0]: 
                    tex=material.texture_slots[0].texture
                    pat=tex.pov
                    texorpig="texture"
                    if pat:
                        uv=""
                        if pat.tex_coords == "UV":
                            uv="uv_mapping"
                        tabWrite("#declare %s_tex = texture { %s\n" %(povMatName,uv))
                        if pat.tex_type == "IMAGE":
                            image_filename = pat.tex_image_path
                            if image_filename:
                                tabWrite("image_pattern {")
                                tabWrite('"%s" \n'%image_filename)
                                tabWrite("gamma %.4g map_type 0 interpolate 2}\n"%pat.tex_ip_gamma)
                        if pat.tex_type == "PATTERN":
                            PovPatterns(pat,tabWrite)
                            PovFunctions(pat,tabWrite,file)
                        if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'}:
                            tabWrite("texture_map {\n")
                        numColor=0
                        for entry in tex.povray_texture:
                            numColor+=1
                            pos=entry.texture_value
                            posColor=string_strip_hyphen(bpy.path.clean_name(entry.name))+"_tex"
                            PovMaps(pat,tabWrite,texorpig,pos,posColor,numColor)
                        if pat.colored_patterns not in {'checker', 'hexagon', 'square', 'triangular', 'brick'}:
                            tabWrite("}\n")
                        if pat.colored_patterns == 'brick':
                            tabWrite("brick_size <%.4g, %.4g, %.4g> mortar %.4g \n"%(pat.brick_size_x, pat.brick_size_y, pat.brick_size_z, pat.brick_mortar))
                        Transform(pat,tabWrite)
                        tabWrite("}\n")
                else:
                    tabWrite("#declare %s_tex = texture{pigment{rgb 0.8}}\n" %povMatName)
    def exportMediaMaterials(material):
            tabWrite("hollow\n")
            tabWrite("interior {\n")
            for t in material.texture_slots:
                if t and t.texture.pov.tex_type == 'PATTERN':
                    if t.texture.pov.tex_use_emis and material.pov.emission_media_ob:
                        tabWrite(" media {\n")
                        tabWrite("  emission %.4g\n"%(material.pov.emission_val))
                        tabWrite("  density {\n")
                        tabWrite("   %s density_map {\n"%(t.texture.pov.colored_patterns))
                        for el in t.texture.color_ramp.elements:
                            pos = el.position
                            col=el.color
                            colR,colG,colB,colA = col[0],col[1],col[2],1-col[3]
                            tabWrite("[%.4g rgb<%.4g,%.4g,%.4g>] \n"%(pos,colR,colG,colB))
                        tabWrite("}}}\n")
                if t and t.texture.pov.tex_type == 'PATTERN':
                    if t.texture.pov.tex_use_absorp and material.pov.absorption_media_ob:
                        tabWrite(" media {\n")
                        tabWrite("  absorption %.4g\n"%(material.pov.absorption_val))
                        tabWrite("  density {\n")
                        tabWrite("   %s density_map {\n"%(t.texture.pov.colored_patterns))
                        for el in t.texture.color_ramp.elements:
                            pos = el.position
                            col=el.color
                            colR,colG,colB,colA = col[0],col[1],col[2],1-col[3]
                            tabWrite("[%.4g rgb<%.4g,%.4g,%.4g>] \n"%(pos,colR,colG,colB))
                        tabWrite("}}}\n")
            if material.pov.scattering_media_ob:
                tabWrite(" media {\n")
                tabWrite("  scattering { %s,\n"%material.pov.s_media_type)
                tabWrite("  <%.4g, %.4g, %.4g>\n"%material.pov.s_media_color[:])
                if material.pov.s_media_type == 5:
                    tabWrite("   eccentricity %.4g \n"%(material.pov.ecc_s_media))
                tabWrite("   extinction %.4g }\n"%(material.pov.ext_s_media))
                tabWrite("  samples %s }\n"%material.pov.s_samples)
            tabWrite("}\n") 

    def exportShapes(scene, ob):
        povName = string_strip_hyphen(bpy.path.clean_name(ob.name))
        tabWrite("#declare %s_ob =\n"%povName)             
        global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
        matrix=global_matrix*ob.matrix_world
        if ob.pov.object_write_type == 'SPHERE':
            tabWrite("sphere{ <0,0,0>,1\n")
        if ob.pov.object_write_type == 'PLANE':
            tabWrite("plane{ <0,0,1>,1\n")
        if ob.pov.object_write_type == 'SUPERELLIPSOID':
            tabWrite("superellipsoid{ <%.4g,%.4g>\n"%(ob.pov.shape_se_param1,ob.pov.shape_se_param2))
        if ob.pov.object_write_type == 'BOX':
            tabWrite("box { <-%.4g,-%.4g,-%.4g>,\n"%ob.scale[:])
            tabWrite("<%.4g,%.4g,%.4g>\n"%ob.scale[:])
        if ob.pov.object_write_type in {'SUPERTORUS'}:
            rMajor = ob.pov.shape_st_radius_major
            rMinor = ob.pov.shape_st_radius_minor
            ring = ob.pov.shape_st_ring
            cross = ob.pov.shape_st_cross
            tabWrite("object{ Supertorus( %.4g,%.4g,%.4g,%.4g,0.001,1.5)\n"%(rMajor,rMinor,ring,cross))
        if ob.pov.object_write_type in {'ROUNDPYRAMID','CONE'} and ob.type=='MESH':
            np=0
            for p in ob.data.polygons:
                if len(p.vertices)==3:
                    np+=1
            sides=np/2
            indxes=[]
            for n in ob.data.vertices:
                indx=n.index
                n=0
                for p in ob.data.polygons:
                    if len(p.vertices)==3:
                        for i in range(0,3):
                            v=p.vertices
                            if v[i]==indx:
                                n+=1            
                                if n==sides:
                                    if i<2:
                                        indxes.append((indx,v[2]))
                                    if i==2:
                                        indxes.append((indx,v[1]))
            if ob.pov.object_write_type in {'ROUNDPYRAMID'}:
                tabWrite("object{ Round_Pyramid_N_out( %s ,\n"%sides)
            if ob.pov.object_write_type in {'CONE'}:
                tabWrite("cone {")
            c_0_indx = indxes[0][0]
            r_0_indx = indxes[0][1]
            c_1_indx = indxes[1][0]
            r_1_indx = indxes[1][1]
            co_c_0 = matrix*ob.data.vertices[c_0_indx].co
            co_r_0 = matrix*ob.data.vertices[r_0_indx].co
            co_c_1 = matrix*ob.data.vertices[c_1_indx].co
            co_r_1 = matrix*ob.data.vertices[r_1_indx].co
            center_0 = "<%.4g,%.4g,%.4g>"%(co_c_0[:])
            radius_0 = sqrt(((co_c_0[0]-co_r_0[0])**2)+((co_c_0[1]-co_r_0[1])**2)+((co_c_0[2]-co_r_0[2])**2))
            tabWrite("%s, %.4g,\n"%(center_0,radius_0))
            center_1 = "<%.4g,%.4g,%.4g>"%(co_c_1[:])
            radius_1 = sqrt(((co_c_1[0]-co_r_1[0])**2)+((co_c_1[1]-co_r_1[1])**2)+((co_c_1[2]-co_r_1[2])**2))
            if ob.pov.object_write_type in {'ROUNDPYRAMID'}:
                tabWrite("%s, %.4g,\n"%(center_1,radius_1))
                filled = 0
                if ob.pov.shape_rp_filled:
                    filled = 1
                merge = 0
                if ob.pov.shape_rp_merge:
                    merge = 1
                tabWrite("%.4g, %s, %s )\n"%(ob.pov.shape_rp_border,filled,merge))
            if ob.pov.object_write_type in {'CONE'}:
                tabWrite("%s, %.4g\n"%(center_1,radius_1))
        if ob.pov.object_write_type == 'HEIGHTFIELD':
            fP="gamma %.4g"%ob.pov.shape_hf_gamma
            if ob.pov.shape_hf_premultiplied:
                fP+=" premultiplied 1"
            if ob.pov.shape_hf_smooth:
                fP+=" smooth"
            fP+=" water_level %.4g"%ob.pov.shape_hf_water
            if ob.pov.shape_hf_hierarchy:
                fP+=" hierarchy 1"
            tabWrite('height_field{"%s" %s\n'%(ob.pov.shape_hf_image,fP))
        if ob.pov.object_write_type == 'SPHERESWEEP':
            points=ob.data.splines[0].points
            numPoints=len(points)
            if ob.data.splines[0].use_cyclic_u:
                numPoints+=3
            tabWrite("sphere_sweep { %s %s,\n"%(ob.pov.shape_ss_type,numPoints))
            if ob.data.splines[0].use_cyclic_u:
                pt1 = points[len(points)-1]
                wpt1 = global_matrix * pt1.co
                tabWrite("<%.4g,%.4g,%.4g>,%.4g,\n" %(wpt1[0], wpt1[1], wpt1[2], pt1.radius*ob.data.bevel_depth))
            numPt=0
            for pt in points:
                numPt+=1
                wpt = global_matrix * pt.co
                if numPt != len(points): 
                    tabWrite("<%.4g,%.4g,%.4g>,%.4g,\n" %(wpt[0], wpt[1], wpt[2], pt.radius*ob.data.bevel_depth))
                if numPt == len(points) and ob.data.splines[0].use_cyclic_u==False: 
                    tabWrite("<%.4g,%.4g,%.4g>,%.4g\n" %(wpt[0], wpt[1], wpt[2], pt.radius*ob.data.bevel_depth))
                if numPt == len(points) and ob.data.splines[0].use_cyclic_u: 
                    tabWrite("<%.4g,%.4g,%.4g>,%.4g,\n" %(wpt[0], wpt[1], wpt[2], pt.radius*ob.data.bevel_depth))
                    for i in range (0,2):
                        endPt=points[i]
                        wpt = global_matrix * endPt.co
                        if i == 0:
                            tabWrite("<%.4g,%.4g,%.4g>,%.4g,\n" %(wpt[0], wpt[1], wpt[2], endPt.radius*ob.data.bevel_depth))
                        if i == 1:
                            tabWrite("<%.4g,%.4g,%.4g>,%.4g\n" %(wpt[0], wpt[1], wpt[2], endPt.radius*ob.data.bevel_depth))
        if ob.pov.object_write_type in {'SOR','LATHE'}:
            splineType=""
            points=None
            lenPts=0
            if ob.pov.object_write_type in {'LATHE'}:
                splineType+=ob.pov.lathe_spline_type
            if ob.pov.object_write_type in {'LATHE'} and ob.pov.lathe_spline_type=='bezier_spline':
                points=ob.data.splines[0].bezier_points
                lenPts=(len(points)-1)*4
            else:
                points=ob.data.splines[0].points
                lenPts=len(points)
            tabWrite("%s { %s %s,\n"%(ob.pov.object_write_type.lower(),splineType,lenPts))
            numPt=0
            if ob.pov.object_write_type in {'LATHE'} and ob.pov.lathe_spline_type=='bezier_spline':
                for i in range(0,len(points)-1):
                    p1=points[i].co
                    pR=points[i].handle_right
                    pL=points[i+1].handle_left
                    p2=points[i+1].co
                    line="<%.4g,%.4g>,"%(p1[0],p1[1])
                    line+="<%.4g,%.4g>,"%(pR[0],pR[1])
                    line+="<%.4g,%.4g>,"%(pL[0],pL[1])
                    if i != len(points)-2:
                        line+="<%.4g,%.4g>,"%(p2[0],p2[1])
                    if i == len(points)-2:
                        line+="<%.4g,%.4g>"%(p2[0],p2[1])
                    tabWrite("%s\n" %line)
            else:
                for pt in points:
                    numPt+=1
                    wpt=pt.co
                    if numPt != len(points): 
                        tabWrite("<%.4g,%.4g>,\n" %(abs(wpt[0]), abs(wpt[1])))
                    if numPt == len(points): 
                        tabWrite("<%.4g,%.4g>\n" %(abs(wpt[0]), abs(wpt[1])))

        if ob.pov.object_write_type == 'CYLINDER':
            point0=ob.data.splines[0].bezier_points[0]
            point1=ob.data.splines[0].bezier_points[1]
            global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
            coord0=global_matrix*point0.co
            co0="<%.4g,"%coord0[0]+"%.4g,"%coord0[1]+"%.4g>"%coord0[2]
            coord1=global_matrix*point1.co
            co1="<%.4g,"%coord1[0]+"%.4g,"%coord1[1]+"%.4g>"%coord1[2]
            radius=ob.data.bevel_depth
            tabWrite("cylinder { %s, %s, %.4g\n"%(co0,co1,radius))
        if ob.pov.object_write_type == 'CONE' and ob.type=='CURVE':
            point0=ob.data.splines[0].bezier_points[0]
            point1=ob.data.splines[0].bezier_points[1]
            global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
            coord0=global_matrix*point0.co
            co0="<%.4g,"%coord0[0]+"%.4g,"%coord0[1]+"%.4g>"%coord0[2]
            coord1=global_matrix*point1.co
            co1="<%.4g,"%coord1[0]+"%.4g,"%coord1[1]+"%.4g>"%coord1[2]
            radius0=ob.data.bevel_depth*point0.radius
            radius1=ob.data.bevel_depth*point1.radius
            tabWrite("cone { %s, %.4g, %s, %.4g\n"%(co0,radius0,co1,radius1))
        if ob.pov.object_write_type == 'TORUS' and ob.type != 'MESH':
            radius_minor=ob.data.bevel_depth
            tabWrite("torus { 1, %.4g\n"%radius_minor)
        exportObjectTextures(scene, ob)
        exportObjectMaterials(scene, ob, povName)
        dS=ob.scale
        if ob.pov.object_write_type not in {'BOX','HEIGHTFIELD','ROUNDPYRAMID','CONE','SPHERE','SOR','LATHE','SUPERELLIPSOID','SUPERTORUS'}:
            tabWrite("scale <%.4g,%.4g,%.4g>\n"%(dS[0],dS[2],dS[1]))
        if ob.pov.object_write_type in {'SPHERE','SOR','LATHE','SUPERELLIPSOID','SUPERTORUS'}:
            tabWrite("scale <%.4g,%.4g,%.4g>\n"%(dS[0],dS[1],dS[2]))
        if ob.pov.object_write_type in {'HEIGHTFIELD'}:
            tabWrite("scale <%.4g,%.4g,%.4g>\n"%(dS[0]*1.5,dS[2]*1.5,dS[1]*1.5))
        if ob.pov.object_write_type not in {'ROUNDPYRAMID','CONE'}:
            rot=tuple([degrees(e) for e in matrix.to_3x3().to_euler()])
            rotX=rot[0]
            if ob.pov.object_write_type in {'HEIGHTFIELD','SPHERESWEEP','CYLINDER','TORUS','SUPERTORUS'}:
                rotX=rot[0]+90 
            tabWrite("rotate <%.4g,%.4g,%.4g>\n"%(rotX,rot[1],rot[2]))
            dT=matrix.translation[:]

            tabWrite("translate <%.6f, %.6f, %.6f>\n"%dT)
        tabWrite("}\n")
        if ob.pov.povobject_visible:
            exportObject(ob, povName)

    def exportShapesObjects(scene, ob):
        povName = string_strip_hyphen(bpy.path.clean_name(ob.name))
        bo=ob.pov.boolean_object_1
        bg=ob.pov.boolean_group_1
        if ob.type=="MESH" and ob.pov.object_write_type not in {'CAIRO','SMB'} and ob.pov.povobject_visible:
            modifiers = ob.modifiers
            isBoolean = 0
            for mod in modifiers:
                if mod.type == 'BOOLEAN' and mod.object:
                    if mod.operation == "UNION":
                        operation = "merge"
                    elif mod.operation == "INTERSECT":
                        operation = "intersection"
                    else:
                        operation = mod.operation.lower()
                    if isBoolean == 0:
                        tabWrite("#declare %s_boolean = %s { \n"%(povName,operation)) 
                        tabWrite("object {%s}\n"%povName)
                        isBoolean += 1
                    child = string_strip_hyphen(bpy.path.clean_name(mod.object.name))
                    tabWrite("object {%s}\n"%child)
            if isBoolean != 0:
                tabWrite("}\n")
                povName=povName+"_boolean"
                exportObject(ob, povName)
        bo2=ob.pov.boolean_object_2
        bg2=ob.pov.boolean_group_2
        bo3=ob.pov.boolean_object_3
        bg3=ob.pov.boolean_group_3
        if ob.pov.povray_boolean and (bo != "" or bg != ""):
            tabWrite("#declare %s_boolean =\n"%povName)
            if bo3 != "" or bg3 != "":
                tabWrite("%s { \n"%ob.pov.boolean_operation_3)
            if bo2 != "" or bg2 != "":
                tabWrite("%s { \n"%ob.pov.boolean_operation_2)
            tabWrite("%s { \n"%ob.pov.boolean_operation_1) 
            tabWrite("object {%s}\n"%povName)
            def writeBooleanObjects(bo,bg):    
                if bo != "":
                    name=string_strip_hyphen(bpy.path.clean_name(bo))
                    tabWrite("object {%s}\n"%name)
                if bg != "":
                    for group in bpy.data.groups:
                        if group.name == bg:
                            for obG in group.objects:
                                obGName = string_strip_hyphen(bpy.path.clean_name(obG.name))
                                tabWrite("object {%s}\n"%obGName)
                tabWrite("}\n")
            writeBooleanObjects(bo,bg)
            bo=bo2
            bg=bg2
            if bo != "" or bg != "":
                writeBooleanObjects(bo,bg)
            bo=bo3
            bg=bg3
            if bo != "" or bg != "":
                writeBooleanObjects(bo,bg)

    def exportBoolean(scene,ob):
        povName = string_strip_hyphen(bpy.path.clean_name(ob.name))
        exportObject(ob, povName)



    def exportNURBS(scene, ob):
        povName = string_strip_hyphen(bpy.path.clean_name(ob.name))
        spline=ob.data.splines[0]
        if spline.type=='NURBS':
            tabWrite("#declare %s = bicubic_patch {\n"%povName)
            tabWrite("type 0 flatness 0.001\n")
            tabWrite("u_steps %s v_steps %s\n"%(spline.point_count_u,spline.point_count_v))
            tabWrite("uv_vectors <0,0> <1,0> <1,1> <0,1>\n")
            global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
            points=spline.points
            numPoint=0
            for point in points:
                numPoint+=1
                value=numPoint%spline.point_count_v
                coord=global_matrix*ob.matrix_world*point.co
                if value != 0:
                    file.write("<%.4g,%.4g,%.4g>,"%(coord[0],coord[1],coord[2]))
                if value == 0:
                    file.write("<%.4g,%.4g,%.4g>\n"%(coord[0],coord[1],coord[2]))
            exportObjectMaterials(scene, ob, povName)
            tabWrite("}\n")
            exportObjectModifiers(ob, povName)



    def exportHair(scene,ob):

        for mod in ob.modifiers:
            if mod.type == 'PARTICLE_SYSTEM':
                if mod.particle_system.settings.type == 'HAIR':
                    modPovName=string_strip_hyphen(bpy.path.clean_name(mod.particle_system.name))
                    fileInc="%s/%s.inc"%(scene.pov.scene_path,modPovName)
                    if os.path.exists(fileInc):
                        tabWrite('#include "%s.inc"\n'%modPovName)
                        tabWrite('object {Hairs%s rotate <-90,0,0>}\n'%modPovName)



    def exportMeshes(scene,ob):
        renderEmitter = True
        global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
        matrix=global_matrix*ob.matrix_world
        if hasattr(ob, 'particle_systems'):
            for pSys in ob.particle_systems:
                if pSys.settings.use_render_emitter:
                    renderEmitter = True
                if pSys.settings.use_render_emitter==False:
                    renderEmitter = False

        if renderEmitter and ob.pov.object_write_type == 'MESH':
            mats = []
            global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
            me = ob.to_mesh(scene, True,'PREVIEW', calc_tessface=False)
            bm = bmesh.new()
            bm.from_mesh(me)
            bm.transform(global_matrix * ob.matrix_world)
            version=bpy.app.version
            if version[1] == 69 and version[2] > 5:
                bmesh.ops.triangulate(bm, faces=bm.faces)
            else:
                bmesh.ops.triangulate(bm, faces=bm.faces, use_beauty=True)
            writeUV=False
            uv_layer = bm.loops.layers.uv.active
            povName = string_strip_hyphen(bpy.path.clean_name(ob.name))
            file=open("%s/%s_dec.inc"%(scene.pov.scene_path,povName),'w')
            file.write("#declare %s_ob =\n"%povName)
            file.write("mesh {\n")
            for face in bm.faces:
                if face.material_index not in mats:
                    mats.append(face.material_index)
            for face in bm.faces:
                if face.smooth:
                    coords="smooth_triangle {"
                else:
                    coords="triangle {"
                verts_in_face = face.verts[:]
                number=0
                uv=" uv_vectors "
                for i in range(0,3):
                    number+=1
                    world_point = face.verts[i].co
                    wp="<%.4g,%.4g,%.4g>"%world_point[:]
                    coords+=wp
                    if face.smooth:
                        vnM=face.verts[i].normal  
                        vn=",<%.4g,%.4g,%.4g>"%vnM[:]
                        coords+=vn
                    if number < 3:
                        coords+=","
                    if uv_layer is not None:
                        uvs="<%.4g,%.4g>"% face.loops[i][uv_layer].uv[:]
                        uv+=uvs
                        if number < 3:
                            uv+=","
                if uv_layer is not None:
                    coords+=uv
                if len(mats)>1:
                    i = face.material_index
                    matSpace = len(ob.material_slots) - (len(mats))
                    mat=ob.material_slots[matSpace + i].material
                    povTexName=string_strip_hyphen(bpy.path.clean_name(mat.name))+"_tex"
                    coords+=" texture {%s}"%povTexName
                coords+="}"
                file.write("%s\n"%coords)
            file.write("}\n")
            file.close()
            tabWrite('#include "%s_dec.inc"\n'%povName)
            bpy.data.meshes.remove(me)
            exportObject(ob, povName)


        if renderEmitter and ob.pov.object_write_type == 'MESH2':
            me = ob.to_mesh(scene, True, 'RENDER')
            global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
            me.transform(global_matrix * ob.matrix_world)
            me_materials = me.materials
            me_faces = me.tessfaces[:]
            povName = string_strip_hyphen(bpy.path.clean_name(ob.name))
            uv_textures = me.tessface_uv_textures
            if len(uv_textures) > 0:
                if me.uv_textures.active and uv_textures.active.data:
                    uv_layer = uv_textures.active.data
            else:
                uv_layer = None

            try:
                vcol_layer = me.vertex_colors.active.data
            except AttributeError:
                vcol_layer = None
            faces_verts = [f.vertices[:] for f in me_faces]
            faces_normals = [(f.normal)[:] for f in me_faces]
            verts_normals = [(v.normal)[:] for v in me.vertices]
            quadCount = sum(1 for f in faces_verts if len(f) == 4)
            file.write("\n")
            tabWrite("#declare %s =\n" % povName)
            tabWrite("mesh2 {\n")
            tabWrite("vertex_vectors {\n")
            tabWrite("%d" % len(me.vertices))

            tabStr = tab * tabLevel
            for v in me.vertices:
                coords=v.co
                if not scene.pov.tempfiles_enable and scene.pov.list_lf_enable:
                    file.write(",\n")
                    file.write(tabStr + "<%.6f, %.6f, %.6f>" % coords[:])
                else:
                    file.write(", ")
                    file.write("<%.6f, %.6f, %.6f>" % coords[:])
            file.write("\n")
            tabWrite("}\n")
            uniqueNormals = {}
            for fi, f in enumerate(me_faces):
                fv = faces_verts[fi]
                if f.use_smooth:
                    for v in fv:
                        key = verts_normals[v]
                        uniqueNormals[key] = [-1]
                else:
                    key = faces_normals[fi]
                    uniqueNormals[key] = [-1]
            tabWrite("normal_vectors {\n")
            tabWrite("%d" % len(uniqueNormals))
            idx = 0
            tabStr = tab * tabLevel
            for no, index in uniqueNormals.items():
                if not scene.pov.tempfiles_enable and scene.pov.list_lf_enable:
                    file.write(",\n")
                    file.write(tabStr + "<%.6f, %.6f, %.6f>" % no)
                else:
                    file.write(", ")
                    file.write("<%.6f, %.6f, %.6f>" % no)
                index[0] = idx
                idx += 1
            file.write("\n")
            
            tabWrite("}\n")
            vertCols = {}

            if uv_layer:
                uniqueUVs = {}
                #n = 0
                for fi, uv in enumerate(uv_layer):
                    if len(faces_verts[fi]) == 4:
                        uvs = uv_layer[fi].uv[0], uv_layer[fi].uv[1], uv_layer[fi].uv[2], uv_layer[fi].uv[3]
                    else:
                        uvs = uv_layer[fi].uv[0], uv_layer[fi].uv[1], uv_layer[fi].uv[2]
                    for uv in uvs:
                        uniqueUVs[uv[:]] = [-1]
                tabWrite("uv_vectors {\n")
                tabWrite("%d" % len(uniqueUVs))
                idx = 0
                tabStr = tab * tabLevel
                for uv, index in uniqueUVs.items():
                    if not scene.pov.tempfiles_enable and scene.pov.list_lf_enable:
                        file.write(",\n")
                        file.write(tabStr + "<%.6f, %.6f>" % uv)
                    else:
                        file.write(", ")
                        file.write("<%.6f, %.6f>" % uv)
                    index[0] = idx
                    idx += 1

                file.write("\n")
                tabWrite("}\n")

            tabWrite("texture_list {\n")
            if len(ob.material_slots)==0:
                tabWrite("1 texture {pigment{rgb 0.8} finish{diffuse 1}}\n")
            if len(ob.material_slots)>0:
                matIndexes = []
                for poly in ob.data.polygons:
                    if poly.material_index not in matIndexes:
                        matIndexes.append(poly.material_index)
                tabWrite("%s\n" % (len(matIndexes))) 
                if len(matIndexes) == 1:
                    tabWrite("texture {%s_tex}\n"%povName)
                if len(matIndexes) > 1:
                    for index in matIndexes:
                        mat=ob.material_slots[index].material
                        if mat:
                            povTexName=string_strip_hyphen(bpy.path.clean_name(mat.name))+"_tex"
                            for tslot in mat.texture_slots:
                                if tslot:
                                    tex=tslot.texture
                                    if tex and tex.pov.tex_use_color:
                                        povTexName=string_strip_hyphen(bpy.path.clean_name(tex.name))
                            tabWrite("texture {%s}\n"%povTexName)
                
            tabWrite("}\n")
            tabWrite("face_indices {\n")
            tabWrite("%d" % (len(me_faces) + quadCount))
            tabStr = tab * tabLevel

            for fi, f in enumerate(me_faces):
                fv = faces_verts[fi]
                material_index = f.material_index
                if len(fv) == 4:
                    indices = (0, 1, 2), (0, 2, 3)
                else:
                    indices = ((0, 1, 2),)

                if vcol_layer:
                    col = vcol_layer[fi]

                    if len(fv) == 4:
                        cols = col.color1, col.color2, col.color3, col.color4
                    else:
                        cols = col.color1, col.color2, col.color3

                if not me_materials or me_materials[material_index] is None:
                    for i1, i2, i3 in indices:
                        if not scene.pov.tempfiles_enable and scene.pov.list_lf_enable:
                            file.write(",\n")
                            # vert count
                            file.write(tabStr + "<%d,%d,%d>,0,0,0" % (fv[i1], fv[i2], fv[i3]))
                        else:
                            file.write(", ")
                            file.write("<%d,%d,%d>" % (fv[i1], fv[i2], fv[i3]))
                else:
                    material = me_materials[material_index]
                    for i1, i2, i3 in indices:
                        if me.vertex_colors and material.use_vertex_color_paint:
                            col1 = cols[i1]
                            col2 = cols[i2]
                            col3 = cols[i3]
                            ci1 = vertCols[col1[0], col1[1], col1[2], material_index][0]
                            ci2 = vertCols[col2[0], col2[1], col2[2], material_index][0]
                            ci3 = vertCols[col3[0], col3[1], col3[2], material_index][0]
                        else:
                            if material.subsurface_scattering.use:
                                diffuse_color = [i * j for i, j in zip(material.subsurface_scattering.color[:], material.diffuse_color[:])]
                            else:
                                diffuse_color = material.diffuse_color[:]
                            ci1 = ci2 = ci3 = f.material_index

                        if not scene.pov.tempfiles_enable and scene.pov.list_lf_enable:
                            file.write(",\n")
                            file.write(tabStr + "<%d,%d,%d>, %d,%d,%d" % \
                                       (fv[i1], fv[i2], fv[i3], ci1, ci2, ci3))
                        else:
                            file.write(", ")
                            file.write("<%d,%d,%d>, %d,%d,%d" % \
                                       (fv[i1], fv[i2], fv[i3], ci1, ci2, ci3))

            file.write("\n")
            tabWrite("}\n")
            tabWrite("normal_indices {\n")
            tabWrite("%d" % (len(me_faces) + quadCount))
            tabStr = tab * tabLevel
            for fi, fv in enumerate(faces_verts):
                if len(fv) == 4:
                    indices = (0, 1, 2), (0, 2, 3)
                else:
                    indices = ((0, 1, 2),)
                for i1, i2, i3 in indices:
                    if me_faces[fi].use_smooth:
                        if not scene.pov.tempfiles_enable and scene.pov.list_lf_enable:
                            file.write(",\n")
                            file.write(tabStr + "<%d,%d,%d>" %\
                            (uniqueNormals[verts_normals[fv[i1]]][0],\
                             uniqueNormals[verts_normals[fv[i2]]][0],\
                             uniqueNormals[verts_normals[fv[i3]]][0]))
                        else:
                            file.write(", ")
                            file.write("<%d,%d,%d>" %\
                            (uniqueNormals[verts_normals[fv[i1]]][0],\
                             uniqueNormals[verts_normals[fv[i2]]][0],\
                             uniqueNormals[verts_normals[fv[i3]]][0]))
                    else:
                        idx = uniqueNormals[faces_normals[fi]][0]
                        if not scene.pov.tempfiles_enable and scene.pov.list_lf_enable:
                            file.write(",\n")
                            file.write(tabStr + "<%d,%d,%d>" % (idx, idx, idx))
                        else:
                            file.write(", ")
                            file.write("<%d,%d,%d>" % (idx, idx, idx))
            file.write("\n")
            tabWrite("}\n")
            if uv_layer:
                tabWrite("uv_indices {\n")
                tabWrite("%d" % (len(me_faces) + quadCount))
                tabStr = tab * tabLevel
                for fi, fv in enumerate(faces_verts):
                    if len(fv) == 4:
                        indices = (0, 1, 2), (0, 2, 3)
                    else:
                        indices = ((0, 1, 2),)
                    uv = uv_layer[fi]
                    if len(faces_verts[fi]) == 4:
                        uvs = uv.uv[0][:], uv.uv[1][:], uv.uv[2][:], uv.uv[3][:]
                    else:
                        uvs = uv.uv[0][:], uv.uv[1][:], uv.uv[2][:]
                    for i1, i2, i3 in indices:
                        if not scene.pov.tempfiles_enable and scene.pov.list_lf_enable:
                            file.write(",\n")
                            file.write(tabStr + "<%d,%d,%d>" % (
                                     uniqueUVs[uvs[i1]][0],\
                                     uniqueUVs[uvs[i2]][0],\
                                     uniqueUVs[uvs[i3]][0]))
                        else:
                            file.write(", ")
                            file.write("<%d,%d,%d>" % (
                                     uniqueUVs[uvs[i1]][0],\
                                     uniqueUVs[uvs[i2]][0],\
                                     uniqueUVs[uvs[i3]][0]))
                file.write("\n")
                tabWrite("}\n")
            bpy.data.meshes.remove(me)
            tabWrite("inside_vector <1,1,1>\n")
            #exportObjectMaterials(scene, ob, povName)
            tabWrite("}\n")
            if ob.pov.povobject_visible:
                exportObject(ob, povName)


    def exportScatteringBox(ob):
        tabWrite("box {\n")
        splines=ob.data.splines
        for spl in splines:
            points=spl.bezier_points
            numPt=0
            for pt in points:
                numPt+=1
                lpt = pt.co
                global_matrix = mathutils.Matrix.Rotation(-pi / 2.0, 4, 'X')
                obMatrix=global_matrix*ob.matrix_world
                wpt = obMatrix * lpt
                if numPt == 1: 
                    tabWrite("<%.4g,%.4g,%.4g>,\n" %(wpt[0], wpt[1], wpt[2]))
                if numPt == 2: 
                    tabWrite("<%.4g,%.4g,%.4g>\n" %(wpt[0], wpt[1], wpt[2]))
        tabWrite("pigment { rgbt 1 } hollow\n")
        tabWrite("photons { pass_through }\n") 
        tabWrite("interior { \n")
        tabWrite(" media {\n")
        tabWrite("  scattering { %s,\n"%ob.pov.smb_type)
        tabWrite("  <%.4g, %.4g, %.4g>\n"%ob.pov.smb_color[:])
        if ob.pov.smb_type == 5:
            tabWrite("eccentricity %.4g\n"%(ob.pov.smb_ecc))
        tabWrite("   extinction %.4g }\n"%(ob.pov.smb_ext))
        tabWrite("  samples %s \n"%ob.pov.smb_samples)
        tabWrite("}}}\n") 





    def exportWorld(world):
        render = scene.render
        camera = scene.camera
        matrix = global_matrix * camera.matrix_world
        if not world:
            return
        if world:
            if not world.use_sky_blend:
                if render.alpha_mode == 'TRANSPARENT':
                    tabWrite("background {rgbt<%.3g, %.3g, %.3g, 0.75>}\n" % \
                             (world.horizon_color[:]))
                elif render.alpha_mode == 'SKY':
                    tabWrite("background {rgbt<%.3g, %.3g, %.3g, 0>}\n" % (world.horizon_color[:]))
            worldTexCount = 0
            for t in world.texture_slots:
                if t:
                    worldTexCount += 1
                if t:
                    image_filename = t.texture.pov.tex_image_path
                    if image_filename != "" and t.use_map_blend:
                        texturesBlend = image_filename
                        t_blend = t
                    if t_blend.texture_coords == 'ANGMAP':
                        mappingBlend = ""
                    else:
                        mappingBlend = " translate <%.4g-0.5,%.4g-0.5,%.4g-0.5> rotate<0,0,0>  " \
                                       "scale <%.4g,%.4g,%.4g>" % \
                                       (t_blend.offset.x / 10.0, t_blend.offset.y / 10.0,
                                        t_blend.offset.z / 10.0, t_blend.scale.x * 0.85,
                                        t_blend.scale.y * 0.85, t_blend.scale.z * 0.85)
                    tabWrite("sky_sphere {\n")
                    tabWrite("pigment {\n")
                    tabWrite("image_map{%s \"%s\" map_type 1 interpolate 2}\n" % \
                             (imageFormat(texturesBlend), texturesBlend))
                    tabWrite("}\n")
                    tabWrite("%s\n" % (mappingBlend))
                    tabWrite("pigment {rgb 0 transmit %s}\n" % (t.texture.intensity))
                    tabWrite("}\n")
            if worldTexCount == 0:
                if world.use_sky_blend:
                    tabWrite("sky_sphere {\n")
                    tabWrite("pigment {\n")
                    tabWrite("gradient y\n")
                    tabWrite("color_map {\n")
                    if render.alpha_mode == 'TRANSPARENT':
                        tabWrite("[0.0 rgbt<%.3g, %.3g, %.3g, 0.99>]\n" % (world.horizon_color[:]))
                        # aa premult not solved with transmit 1
                        tabWrite("[0.2 rgbt<%.3g, %.3g, %.3g, 0.99>]\n" % (world.zenith_color[:]))
                    else:
                        tabWrite("[0.0 rgbt<%.3g, %.3g, %.3g, 0>]\n" % (world.horizon_color[:]))
                        tabWrite("[0.2 rgbt<%.3g, %.3g, %.3g, 0>]\n" % (world.zenith_color[:]))
                    tabWrite("}\n")
                    tabWrite("}\n")
                    tabWrite("}\n")

        ###############################################################

        mist = world.mist_settings

        if world.pov.fog_use:
            for wp in world.povray_fogs:
                tabWrite("fog {\n")
                tabWrite("fog_type 2\n")
                tabWrite("color rgb<%.3g, %.3g, %.3g>\n" %(wp.fog_color[:]))
                tabWrite("distance %.4g\n" %(wp.fog_distance))
                tabWrite("up <%.3g, %.3g, %.3g>\n" %(wp.fog_up_x,wp.fog_up_y,wp.fog_up_z))
                tabWrite("fog_offset %.2f\n" %wp.fog_offset)
                tabWrite("fog_alt %s\n"%wp.fog_alt)
                tabWrite("turbulence <%.3g, %.3g, %.3g>\n" %(wp.fog_turb_x,wp.fog_turb_y,wp.fog_turb_z))
                tabWrite("turb_depth %s\n"%wp.fog_turb_depth)
                tabWrite("lambda %s\n"%wp.fog_lambda)
                tabWrite("octaves %s\n"%wp.fog_octaves)
                tabWrite("omega %s\n"%wp.fog_omega)
                tabWrite("}\n")
        if scene.pov.media_enable:
            tabWrite("media {\n")
            tabWrite("scattering { 1, rgb <%.4g, %.4g, %.4g>}\n" % scene.pov.media_color[:])
            tabWrite("samples %.d\n" % scene.pov.media_samples)
            tabWrite("}\n")

    def exportGlobalSettings(scene):

        tabWrite("global_settings {\n")
        tabWrite("assumed_gamma %s\n"% scene.pov.assumed_gamma)
        tabWrite("max_trace_level %d\n" % scene.pov.max_trace_level)
        if scene.pov.mm_per_unit != 10.0:
            tabWrite("mm_per_unit %.3g\n" % scene.pov.mm_per_unit)
        if scene.pov.sss_enable:
            line=""
            if scene.pov.sss_samples_enable:
                line+="samples %s, %s"%(scene.pov.sss_samples_max,scene.pov.sss_samples_min)
            if scene.pov.sss_radiosity:
                line+=" radiosity on"
            tabWrite("subsurface {%s}\n" % line)



        if scene.pov.radio_enable:
            tabWrite("radiosity {\n")
            tabWrite("adc_bailout %.4g\n" % scene.pov.radio_adc_bailout)
            tabWrite("always_sample %d\n" % scene.pov.radio_always_sample)
            tabWrite("brightness %.4g\n" % scene.pov.radio_brightness)
            tabWrite("count %d\n" % scene.pov.radio_count)
            tabWrite("error_bound %.4g\n" % scene.pov.radio_error_bound)
            tabWrite("gray_threshold %.4g\n" % scene.pov.radio_gray_threshold)
            tabWrite("low_error_factor %.4g\n" % scene.pov.radio_low_error_factor)
            tabWrite("media %d\n" % scene.pov.radio_media)
            tabWrite("minimum_reuse %.4g\n" % scene.pov.radio_minimum_reuse)
            tabWrite("nearest_count %d\n" % scene.pov.radio_nearest_count)
            tabWrite("normal %d\n" % scene.pov.radio_normal)
            tabWrite("pretrace_start %.3g\n" % scene.pov.radio_pretrace_start)
            tabWrite("pretrace_end %.3g\n" % scene.pov.radio_pretrace_end)
            tabWrite("recursion_limit %d\n" % scene.pov.radio_recursion_limit)
            tabWrite("}\n")
        onceSss = 1


        for material in bpy.data.materials:
            if material.subsurface_scattering.use and onceSss:
                tabWrite("mm_per_unit %.6f\n" % \
                         (material.subsurface_scattering.scale * 10000.0))
                sslt_samples = (11 - material.subsurface_scattering.error_threshold) * 10
                tabWrite("subsurface { samples %d, %d }\n" % (sslt_samples, sslt_samples / 10))
                onceSss = 0


        if scene.pov.photon_enable:
            tabWrite("photons {\n")
            tabWrite("count %d\n" % scene.pov.photon_count)
            if scene.pov.photon_media_enable:
                tabWrite("media %d, %.4g\n" % (scene.pov.photon_media_steps,scene.pov.photon_media_factor))
            if scene.pov.photon_advanced:
                if scene.pov.photon_adc:
                    tabWrite("adc_bailout %.3g\n" % scene.pov.photon_adc_bailout)
                if scene.pov.photon_gather:
                    tabWrite("gather %d, %d\n" % (scene.pov.photon_gather_min, scene.pov.photon_gather_max))
                if scene.pov.photon_autostop:
                    tabWrite("autostop %d\n" % scene.pov.photon_autostop_value)
            tabWrite("}\n")


        tabWrite("}\n")

    sel = renderable_objects(scene)
    file.write("#version 3.7;\n")
    file.write('#include "functions.inc"\n')
    file.write('#include "shapes3.inc"\n')
    exportGlobalSettings(scene)
    exportWorld(scene.world)
    declare_materials=[]
    for ob in sel:
        for slot in ob.material_slots:
            material = slot.material
            if material and material.pov.povray_mat_type != "PATTERN" and material.name not in declare_materials:
                declare_materials.append(material.name)
                declareFinPigNor(material)
    declare_materials=[]
    for ob in sel:
        for slot in ob.material_slots:
            material = slot.material
            if material and material.name not in declare_materials:
                declare_materials.append(material.name)
                declareTexture(material)
    declare_materials=[]
    for ob in sel:
        numMat=0
        for slot in ob.material_slots:
            material = slot.material
            if material and material.pov.povray_mat_type == "PATTERN" and material.name not in declare_materials:
                declare_materials.append(material.name)
                declareFinPigNor(material)
    for ob in sel:
        if ob.type == 'MESH' and ob.pov.object_write_type in {'MESH','MESH2'}:
            exportMeshes(scene,ob)
            exportHair(scene,ob)
        if (ob.type == 'MESH' or ob.type == 'CURVE') and ob.pov.object_write_type not in {'MESH','MESH2'}:
            exportShapes(scene,ob)




    exportLamps([l for l in sel if l.type == 'LAMP'])
    exportCamera()
    file.close()

def write_pov_ini(scene, filename_ini, filename_pov, filename_image):
    render = scene.render
    workDir=os.path.dirname(__file__)
    includeDir=('%s/include/'%workDir)
    if sys.platform[:3] == "win":
        includeDir="C:\\povray\\include\\"
    x = int(render.resolution_x * render.resolution_percentage * 0.01)
    y = int(render.resolution_y * render.resolution_percentage * 0.01)

    file = open(filename_ini, "w")
    file.write("Version=3.7\n")
    file.write("Input_File_Name='%s'\n" % filename_pov)
    file.write("Output_File_Name='%s'\n" % filename_image)
    file.write('Library_Path="%s"\n'%scene.pov.scene_path)
    file.write("Width=%d\n" % x)
    file.write("Height=%d\n" % y)

    # Border render.
    if render.use_border:
        file.write("Start_Column=%4g\n" % render.border_min_x)
        file.write("End_Column=%4g\n" % (render.border_max_x))

        file.write("Start_Row=%4g\n" % (1.0 - render.border_max_y))
        file.write("End_Row=%4g\n" % (1.0 - render.border_min_y))

    file.write("Bounding_Method=2\n")  # The new automatic BSP is faster in most scenes

    # Activated (turn this back off when better live exchange is done between the two programs
    # (see next comment)
    file.write("Display=1\n")
    file.write("Display_Gamma=%.4g\n"%scene.pov.display_gamma)
    file.write("Pause_When_Done=0\n")
    file.write("Output_File_Type=N\n")
    file.write("Output_Alpha=1\n")

    if scene.pov.antialias_enable:
        file.write("Antialias=on\n")
        file.write("Sampling_Method=%s\n" % scene.pov.antialias_method)
        file.write("Antialias_Depth=%d\n" % scene.pov.antialias_depth)
        file.write("Antialias_Threshold=%.3g\n" % scene.pov.antialias_threshold)
        file.write("Antialias_Gamma=%.3g\n" % scene.pov.antialias_gamma)
        if scene.pov.jitter_enable:
            file.write("Jitter=on\n")
            file.write("Jitter_Amount=%3g\n" % scene.pov.jitter_amount)
        else:
            file.write("Jitter=off\n")
    else:
        file.write("Antialias=off\n")
    file.close()


class PovrayRender(bpy.types.RenderEngine):
    bl_idname = 'POVRAY_RENDER'
    bl_label = "POV-Ray 3.7"
    DELAY = 0.5

    @staticmethod
    def _locate_binary():
        addon_prefs = bpy.context.user_preferences.addons[__package__].preferences
        pov_binary = addon_prefs.filepath_povray
        if pov_binary:
            if os.path.exists(pov_binary):
                return pov_binary
            else:
                print("User Preference to povray %r NOT FOUND, checking $PATH" % pov_binary)
        if sys.platform[:3] == "win":
            import winreg
            win_reg_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, "Software\\POV-Ray\\v3.7\\Windows")
            win_home = winreg.QueryValueEx(win_reg_key, "Home")[0]
            pov_binary = os.path.join(win_home, "bin", "pvengine64.exe")
            if os.path.exists(pov_binary):
                return pov_binary
            pov_binary = os.path.join(win_home, "bin", "pvengine.exe")
            if os.path.exists(pov_binary):
                return pov_binary
        pov_binary_default = "povray"
        os_path_ls = os.getenv("PATH").split(':') + [""]
        for dir_name in os_path_ls:
            pov_binary = os.path.join(dir_name, pov_binary_default)
            if os.path.exists(pov_binary):
                return pov_binary
        return ""

    def _export(self, scene, povPath, renderImagePath):
        import tempfile

        if scene.pov.tempfiles_enable:
            self._temp_file_in = tempfile.NamedTemporaryFile(suffix=".pov", delete=False).name
            self._temp_file_out = tempfile.NamedTemporaryFile(suffix=".png", delete=False).name
            self._temp_file_ini = tempfile.NamedTemporaryFile(suffix=".ini", delete=False).name
        else:
            self._temp_file_in = povPath + ".pov"
            self._temp_file_out = renderImagePath + ".png"
            self._temp_file_ini = povPath + ".ini"

        def info_callback(txt):
            self.update_stats("", "POV-Ray 3.7: " + txt)

        write_pov(self._temp_file_in, scene, info_callback)

    def _render(self, scene):
        try:
            os.remove(self._temp_file_out)
        except OSError:
            pass

        pov_binary = PovrayRender._locate_binary()
        if not pov_binary:
            print("POV-Ray 3.7: could not execute povray, possibly POV-Ray isn't installed")
            return False

        write_pov_ini(scene, self._temp_file_ini, self._temp_file_in, self._temp_file_out)

        print ("***-STARTING-***")

        extra_args = []

        if scene.pov.command_line_switches != "":
            for newArg in scene.pov.command_line_switches.split(" "):
                extra_args.append(newArg)

        self._is_windows = False
        if sys.platform[:3] == "win":
            self._is_windows = True

        try:
            env = {'POV_DISPLAY_SCALED': 'off'}
            env.update(os.environ)
            self._process = subprocess.Popen([pov_binary, self._temp_file_ini],
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                env=env)
        except OSError:
            print("POV-Ray 3.7: could not execute '%s'" % pov_binary)
            import traceback
            traceback.print_exc()
            print ("***-DONE-***")
            return False

        else:
            print("POV-Ray 3.7 found")
            print("Command line arguments passed: " + str(extra_args))
            return True

    def _cleanup(self):
        for f in (self._temp_file_in, self._temp_file_ini, self._temp_file_out):
            for i in range(5):
                try:
                    os.unlink(f)
                    break
                except OSError:
                    time.sleep(self.DELAY)

    def render(self, scene):
        import tempfile
        print("***INITIALIZING***")
        blendSceneName = bpy.data.filepath.split(os.path.sep)[-1].split(".")[0]
        povSceneName = ""
        povPath = ""
        renderImagePath = ""
        scene.frame_set(scene.frame_current)
        if not scene.pov.tempfiles_enable:
            povPath = bpy.path.abspath(scene.pov.scene_path).replace('\\', '/')
            if povPath == "":
                if bpy.data.is_saved:
                    povPath = bpy.path.abspath("//")
                else:
                    povPath = tempfile.gettempdir()
            elif povPath.endswith("/"):
                if povPath == "/":
                    povPath = bpy.path.abspath("//")
                else:
                    povPath = bpy.path.abspath(scene.pov.scene_path)

            if not os.path.exists(povPath):
                try:
                    os.makedirs(povPath)
                except:
                    import traceback
                    traceback.print_exc()

                    print("POV-Ray 3.7: Cannot create scenes directory: %r" % povPath)
                    self.update_stats("", "POV-Ray 3.7: Cannot create scenes directory %r" % \
                                      povPath)
                    time.sleep(2.0)
                    return
            if scene.pov.scene_name == "":
                if blendSceneName != "":
                    povSceneName = blendSceneName
                else:
                    povSceneName = "untitled"
            else:
                povSceneName = scene.pov.scene_name
                if os.path.isfile(povSceneName):
                    povSceneName = os.path.basename(povSceneName)
                povSceneName = povSceneName.split('/')[-1].split('\\')[-1]
                if not povSceneName:
                    print("POV-Ray 3.7: Invalid scene name")
                    self.update_stats("", "POV-Ray 3.7: Invalid scene name")
                    time.sleep(2.0)
                    return
                povSceneName = os.path.splitext(povSceneName)[0]

            print("Scene name: " + povSceneName)
            print("Export path: " + povPath)
            povPath = os.path.join(povPath, povSceneName)
            povPath = os.path.realpath(povPath)
            renderImagePath = povPath
        self.update_stats("", "POV-Ray 3.7: Exporting data from Blender")
        self._export(scene, povPath, renderImagePath)
        self.update_stats("", "POV-Ray 3.7: Parsing File")
        if not self._render(scene):
            self.update_stats("", "POV-Ray 3.7: Not found")
            return
        r = scene.render
        x = int(r.resolution_x * r.resolution_percentage * 0.01)
        y = int(r.resolution_y * r.resolution_percentage * 0.01)
        def _test_wait():
            time.sleep(self.DELAY)
            if self.test_break():
                try:
                    self._process.terminate()
                    print("***POV INTERRUPTED***")
                except OSError:
                    pass
                return False

            poll_result = self._process.poll()
            if poll_result is not None:
                if poll_result < 0:
                    print("***POV PROCESS FAILED : %s ***" % poll_result)
                    self.update_stats("", "POV-Ray 3.7: Failed")
                return False

            return True
        parsing = re.compile(br"= \[Parsing\.\.\.\] =")
        rendering = re.compile(br"= \[Rendering\.\.\.\] =")
        percent = re.compile(r"\(([0-9]{1,3})%\)")
        data = b""
        last_line = ""
        while _test_wait():
            if self._is_windows:
                self.update_stats("", "POV-Ray 3.7: Rendering File")
            else:
                t_data = self._process.stdout.read(10000)
                if not t_data:
                    continue
                data += t_data
                t_data = str(t_data).replace('\\r\\n', '\\n').replace('\\r', '\r')
                lines = t_data.split('\\n')
                last_line += lines[0]
                lines[0] = last_line
                print('\n'.join(lines), end="")
                last_line = lines[-1]
                if rendering.search(data):
                    _pov_rendering = True
                    match = percent.findall(str(data))
                    if match:
                        self.update_stats("", "POV-Ray 3.7: Rendering File (%s%%)" % match[-1])
                    else:
                        self.update_stats("", "POV-Ray 3.7: Rendering File")
                elif parsing.search(data):
                    self.update_stats("", "POV-Ray 3.7: Parsing File")
        if os.path.exists(self._temp_file_out):
            xmin = int(r.border_min_x * x)
            ymin = int(r.border_min_y * y)
            xmax = int(r.border_max_x * x)
            ymax = int(r.border_max_y * y)
            result = self.begin_result(0, 0, x, y)
            lay = result.layers[0]
            time.sleep(self.DELAY)
            try:
                lay.load_from_file(self._temp_file_out)
            except RuntimeError:
                print("***POV ERROR WHILE READING OUTPUT FILE***")
            self.end_result(result)
        else:
            print("***POV FILE NOT FOUND***")
        print("***POV FINISHED***")
        self.update_stats("", "")
        if scene.pov.tempfiles_enable or scene.pov.deletefiles_enable:
            self._cleanup()
